// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <algorithm>
#include <atomic>
#include <assert.h>
#include <cctype>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <poll.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "error.h"
#include "http_server.h"
#include "picio.h"
#include "source.h"
#include "utils.h"
#include "log.h"
#include "motion_trigger.h"
#include "target.h"
#include "target_avi.h"
#include "target_ffmpeg.h"
#include "cairo.h"
#include "http_server_content.h"
#include "http_server_rest.h"
#include "http_server_support.h"

typedef struct {
	int fd;
	double fps;
	int quality;
	int time_limit;
	const std::vector<filter *> *filters;
	std::atomic_bool *global_stopflag;
	resize *r;
	int resize_w, resize_h;
	source *motion_compatible;
	bool allow_admin, archive_acces, is_rest;
	configuration_t *cfg;
	std::string snapshot_dir;
	instance_t *views, *limit_to;
	bool use_previous_when_fail;
	http_auth *auth;
} http_thread_t;

bool sort_files_last_change(const file_t &left, const file_t & right)
{
	return left.last_change < right.last_change;
}

bool sort_files_size(const file_t &left, const file_t & right)
{
	return left.size < right.size;
}

bool sort_files_name(const file_t &left, const file_t & right)
{
	return left.name.compare(right.name) < 0;
}

const std::string http_200_header =
		"HTTP/1.0 200\r\n"
		"Cache-Control: no-cache\r\n"
		"Pragma: no-cache\r\n"
		"Server: " NAME " " VERSION "\r\n"
		"Expires: thu, 01 dec 1994 16:00:00 gmt\r\n"
		"Content-Type: text/html\r\n"
		"Connection: close\r\n"
		"\r\n";

const std::string http_401_header =
		"HTTP/1.0 401\r\n"
		"WWW-Authenticate: Basic realm=\"Please authenticate\"\r\n"
		"Cache-Control: no-cache\r\n"
		"Pragma: no-cache\r\n"
		"Server: " NAME " " VERSION "\r\n"
		"Expires: thu, 01 dec 1994 16:00:00 gmt\r\n"
		"Content-Type: text/html\r\n"
		"Connection: close\r\n"
		"\r\n";

const std::string html_header =
		"<!DOCTYPE html>"
		"<html>"
		"<head>"
		"<link href=\"stylesheet.css\" rel=\"stylesheet\" media=\"screen\">"
		"<link rel=\"shortcut icon\" href=\"favicon.ico\">"
		"<title>" NAME " " VERSION "</title>"
		"</head>"
		"<body>"
		"<h1>" NAME " " VERSION "%s</h1>";

const std::string html_tail =
		"<p><br><br><br></p><hr><div id=\"tail\"><p>" NAME " was written by folkert@vanheusden.com</p></div>"
		"</body>"
		"</html>";

const std::string action_failed = "<h2>%s</h2><p>Action failed</p>";
const std::string action_succeeded = "<h2>%s</h2><p>Action succeeded</p>";

void http_server::mjpeg_stream_url(configuration_t *const cfg, const std::string & id, std::string *const img_url, std::string *const page_url)
{
	instance_t *inst = nullptr;
	interface *i = nullptr;
	find_by_id(cfg, id, &inst, &i);

	if (!inst) {
		*img_url = *page_url = "id unknown";
		return;
	}

	*img_url = myformat("stream.mjpeg?inst=%s", url_escape(inst -> name).c_str()); 
	*page_url = myformat("index.html?inst=%s", url_escape(inst -> name).c_str()); 
}

bool http_draw_histogram(const std::string & font, const int x_off, const int y_off, const int width, const int height, std::vector<std::string> & labels_x, const double *const val_y, const int n_values, const std::string & metaStr, const int fd)
{
	std::string header = 
		"HTTP/1.0 200\r\n"
		"Cache-Control: no-cache\r\n"
		"Pragma: no-cache\r\n"
		"Server: " NAME " " VERSION "\r\n"
		"Expires: thu, 01 dec 1994 16:00:00 gmt\r\n"
		"Content-Type: image/png\r\n"
		"Connection: close\r\n"
		"\r\n";

	if (WRITE(fd, header.c_str(), header.size()) == -1)
		return false;

	uint8_t *temp = nullptr;
	size_t temp_len = 0;
	FILE *fh = open_memstream((char **)&temp, &temp_len);
	draw_histogram(font, x_off, y_off, width, height, labels_x, val_y, n_values, metaStr, fh);
	fclose(fh);

	bool rc = WRITE(fd, (const char *)temp, temp_len) != -1;

	free(temp);

	return rc;
}

void handle_http_client(int cfd, double fps, int quality, int time_limit, const std::vector<filter *> *const filters, std::atomic_bool *const global_stopflag, resize *const r, const int resize_w, const int resize_h, source *const motion_compatible, const std::string & snapshot_dir, const bool allow_admin, const bool archive_acces, configuration_t *const cfg, instance_t *const limit_to, instance_t *const views, const bool use_previous_when_fail, http_auth *const auth)
{
	sigset_t all_sigs;
	sigfillset(&all_sigs);
	pthread_sigmask(SIG_BLOCK, &all_sigs, nullptr);

	char request_headers[65536] = { 0 };
	int h_n = 0;

	struct pollfd fds[1] { { cfd, POLLIN, 0 } };

	for(;!*global_stopflag && !motion_compatible;) {
		if (poll(fds, 1, 250) == 0)
			continue;

		int rc = read(cfd, &request_headers[h_n], sizeof request_headers - h_n - 1);
		if (rc == -1)
		{
			if (errno == EINTR || errno == EAGAIN) {
				log(LL_DEBUG, "EINTR/EAGAIN %d", errno);
				continue;
			}

			log(LL_DEBUG, "error receiving request headers");
			close(cfd);
			return;
		}
		else if (rc == 0)
		{
			log(LL_DEBUG, "error receiving request headers");
			close(cfd);
			return;
		}

		h_n += rc;

		request_headers[h_n] = 0x00;

		if (strstr(request_headers, "\r\n\r\n") || strstr(request_headers, "\n\n"))
			break;
	}

	char *Path = nullptr;

	bool get = true;
	if (motion_compatible) {
	}
	else if (strncmp(request_headers, "GET ", 4) == 0)
	{
		get = true;
		Path = strdup(&request_headers[4]);
	}
	else if (strncmp(request_headers, "HEAD ", 5) == 0)
	{
		get = false;
		Path = strdup(&request_headers[5]);
	}
	else
	{
		close(cfd);
		return;
	}

	char *unescaped_path = motion_compatible ? strdup("/") : un_url_escape(Path);

	char *path = unescaped_path;
	if (!motion_compatible) {
		while(*path == '/')
			path++;
	}

	std::map<std::string, std::string> pars;

	instance_t *inst = limit_to;
	source *s = nullptr;
	bool auth_ok = true;

	if (!motion_compatible) {
		char *dummy = strchr(path, '\r');
		if (!dummy)
			dummy = strchr(path, '\n');
		if (dummy)
			*dummy = 0x00;

		dummy = strrchr(path, ' ');
		if (dummy)
			*dummy = 0x00;

		std::string ff;
		char *forwarded_for = strcasestr(request_headers, "X-Forwarded-For:");
		if (forwarded_for) {
			forwarded_for = strdup(forwarded_for);

			dummy = strchr(forwarded_for, '\r');
			if (dummy)
				*dummy = 0x00;
			dummy = strchr(forwarded_for, '\n');
			if (dummy)
				*dummy = 0x00;

			ff = myformat(" (%s)", &forwarded_for[17]);

			free(forwarded_for);
		}

		if (auth) {
			char *authorization = strcasestr(request_headers, "Authorization:");
			if (authorization) {
				authorization = strdup(authorization);

				dummy = strchr(authorization, '\r');
				if (dummy)
					*dummy = 0x00;
				dummy = strchr(authorization, '\n');
				if (dummy)
					*dummy = 0x00;

				std::vector<std::string> *parts = split(&authorization[15], " ");
				if (parts->size() != 2)
					auth_ok = false;
				else if (str_tolower(parts->at(0)) != "basic")
					auth_ok = false;
				else {
					std::string decoded = base64_decode(parts->at(1));
					size_t colon = decoded.find(":");
					if (colon == std::string::npos)
						auth_ok = false;
					else {
						std::string username = decoded.substr(0, colon);
						std::string password = decoded.substr(colon + 1);

						auto result = auth->authenticate(username, password);
						if (!std::get<0>(result)) {
							log(LL_WARNING, "HTTP authentication failed: %s", std::get<1>(result).c_str());
							auth_ok = false;
						}
					}
				}
				delete parts;

				free(authorization);
			}
			else {
				auth_ok = false;
			}
		}

		log(LL_INFO, "URL: %s%s", path, ff.c_str());

		char *q = strchr(path, '?');
		if (q) {
			*q = 0x00;

			char *p = q + 1;
			while(p) {
				std::string pair;
				char *end = strchr(p, '&');
				if (end) {
					pair = std::string(p, end - p);
					p = end + 1;
				}
				else {
					pair = std::string(p);
					p = nullptr;
				}

				std::string key, value;

				if (ssize_t is = pair.find('='); is == std::string::npos)
					key = pair;
				else {
					key = pair.substr(0, is);
					value = pair.substr(is + 1);
				}

				pars.insert(std::pair<std::string, std::string>(key, value));
			}

			if (auto int_it = pars.find("int"); int_it != pars.end())
				find_by_id(cfg, int_it -> second, &inst, (interface **)&s);

			if (auto inst_it = pars.find("inst"); inst_it != pars.end())
				inst = find_instance_by_name(cfg, inst_it -> second);

			if (!s)
				s = find_source(inst);
		}
	}

	if (s == nullptr && motion_compatible)
		s = motion_compatible;

	if (s)
		s -> register_user();

	auto vp_it = pars.find("view-proxy"); // interface
	const bool is_view_proxy = vp_it != pars.end();

	// iup = instance url parameter
	const std::string iup = inst ? myformat("?inst=%s", url_escape(inst -> name).c_str()) : "";

	//if (inst)
	//printf("%s\n", inst -> name.c_str());

	const std::string page_header = myformat(html_header.c_str(), inst ? (" / " + inst -> name).c_str() : "");

	if (!auth_ok) {
		if (WRITE(cfd, http_401_header.c_str(), http_401_header.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if ((path == nullptr || strcmp(path, "stream.mjpeg") == 0 || motion_compatible) && s)
		send_mjpeg_stream(cfd, s, fps, quality, get, time_limit, filters, global_stopflag, r, resize_w, resize_h, cfg, is_view_proxy, use_previous_when_fail);
	else if (strcmp(path, "stream.mpng") == 0 && s)
		send_mpng_stream(cfd, s, fps, get, time_limit, filters, global_stopflag, r, resize_w, resize_h, cfg, is_view_proxy, use_previous_when_fail);
	else if (strcmp(path, "image.png") == 0 && s)
		send_png_frame(cfd, s, get, filters, r, resize_w, resize_h, cfg, is_view_proxy, use_previous_when_fail);
	else if (strcmp(path, "image.jpg") == 0 && s)
		send_jpg_frame(cfd, s, get, quality, filters, r, resize_w, resize_h, cfg, is_view_proxy, use_previous_when_fail);
	else if (strcmp(path, "hist") == 0) {
		std::vector<std::string> labels_x;
//		labels_x.push_back("een");
//		labels_x.push_back("twee");
//		labels_x.push_back("drie");
		double val_y[] = { 1, 2, 3 };
		http_draw_histogram("Arial", 0, 0, 800, 600, labels_x, val_y, 3, "test", cfd);
	}
	else if (strcmp(path, "stylesheet.css") == 0) {
		struct stat st;
		if (stat("stylesheet.css", &st) == 0)
			send_file(cfd, "./", "stylesheet.css", false);
		else {
			std::string reply = http_200_header + "\r\n";

			if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
				log(LL_DEBUG, "short write on response");
		}
	}
	else if (strcmp(path, "stream.html") == 0)
	{
		std::string reply = http_200_header + page_header + myformat("<img src=\"stream.mjpeg%s\">", iup.c_str()) + html_tail;

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if (strcmp(path, "index.html") == 0 || strcmp(path, "") == 0)
	{
		std::string reply = "???";

		if (inst) {
			reply = http_200_header + page_header + "<div id=\"main\">"
				"<ul>" +
				myformat("<li><a href=\"stream.mjpeg%s\">MJPEG stream</a>"
						"<li><a href=\"stream.html%s\">Same MJPEG stream but in a HTML wrapper</a>"
						"<li><a href=\"stream.mpng%s\">MPNG stream</a>"
						"<li><a href=\"image.jpg%s\">Show snapshot in JPG format</a>"
						"<li><a href=\"image.png%s\">Show snapshot in PNG format</a>", iup.c_str(), iup.c_str(), iup.c_str(), iup.c_str(), iup.c_str());

			if (allow_admin) {
				reply += myformat("<li><a href=\"snapshot-img/%s\">Take a snapshot and store it on disk</a>"
						"<li><a href=\"snapshot-video/%s\">Start a video recording</a>", iup.c_str(), iup.c_str());
			}

			if (archive_acces) {
				reply += "<li><a href=\"view-snapshots/\">View recordings</a>";
			}

			reply += "</ul></div>";

			if (allow_admin) {
				cfg -> lock.lock();

				for(interface *i : inst -> interfaces)
					reply += describe_interface(inst, i);
				cfg -> lock.unlock();
			}

			reply += html_tail; 
		}
		else {
			reply = http_200_header + page_header + "<h2>views</h2>";

			if (!views || views -> interfaces.empty())
				reply += "<b>No view defined</b>";
			else {
				reply += "<ul>";

				for(interface *v : views->interfaces)  {
					if (v->get_class_type() != CT_VIEW)
						continue;

					view *cv = (view *)v;
					reply += myformat("<li><a href=\"view-view?id=%s\">%s</a>", cv -> get_id().c_str(), (cv -> get_descr().empty() ? cv -> get_id() : cv -> get_descr()).c_str());
				}

				reply += "</ul>";
			}

			reply += "<h2>instances</h2><ul>";
			for(instance_t *inst : cfg -> instances) {
				if (inst -> name == "views")
					continue;

				reply += myformat("<li><a href=\"index.html?inst=%s\">%s</a>", url_escape(inst -> name).c_str(), inst -> name.c_str());
			}
			reply += "</ul>";
		}

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if ((strncmp(path, "view-snapshots/send-file", 24) == 0 || strncmp(path, "send-file", 9) == 0) && (archive_acces || allow_admin)) {
		std::string file;

		if (auto file_it = pars.find("file"); file_it != pars.end())
			file = file_it -> second;

		bool force_dl = pars.find("download") != pars.end();

		if (!file.empty() && validate_file(snapshot_dir, file)) {
			send_file(cfd, snapshot_dir, file.c_str(), force_dl);
		}
		else {
			log(LL_WARNING, "%s is not a valid file", file.c_str());

			std::string reply = "HTTP/1.0 404\r\n\r\nfailed";

			if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
				log(LL_DEBUG, "short write on response header");
		}
	}
	else if (strcmp(path, "view-view") == 0 && views) {
		std::string id;

		if (auto view_it = pars.find("id"); view_it != pars.end())
			id = view_it -> second;

		view *v = find_view(views, id);

		std::string reply;
		if (v)
			reply = http_200_header + v -> get_html(pars);

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response");
	}
	else if ((strncmp(path, "view-snapshots/view-file", 24) == 0 || strncmp(path, "view-file", 9) == 0) && (archive_acces || allow_admin)) {
		std::string file;

		if (auto file_it = pars.find("file"); file_it != pars.end())
			file = file_it -> second;

		std::string reply = http_200_header + page_header + 
			myformat("<video controls autoplay><source src=\"send-file?file=%s\">Your browser does not support the video tag.</video>", file.c_str())
			+ html_tail;

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response");
	}
	else if (strcmp(path, "view-snapshots/") == 0 && (archive_acces || allow_admin)) {
		std::string reply = http_200_header + page_header + "<h2>list of files in " + snapshot_dir + "</h2><table border=1>";

		auto *files = load_filelist(snapshot_dir, "");

		std::string order;

		if (auto order_it = pars.find("order"); order_it != pars.end())
			order = order_it -> second;

		if (order == "" || order == "last-change")
			std::sort(files -> begin(), files -> end(), sort_files_last_change);
		else if (order == "name")
			std::sort(files -> begin(), files -> end(), sort_files_name);
		else if (order == "size")
			std::sort(files -> begin(), files -> end(), sort_files_size);
		else
			reply += myformat("Ignoring unknown sort method %s", order.c_str());

		reply += "<tr><th><a href=\"?order=name\">name</a></th><th><a href=\"?order=last-change\">last change</a></th><th><a href=\"?order=size\">size</a></th><th></th></tr>";

		for(auto file : *files) {
			reply += std::string("<tr><td><a href=\"send-file?download&file=") + file.name + "\">" + file.name + "</a></td><td>" + myctime(file.last_change) + "</td><td>" + myformat("%zu", file.size) + "</td><td>" + myformat("<a href=\"view-file?file=%s\">view</a>", file.name.c_str()) + "</td></tr>";
		}

		delete files;

		reply += "</table>" + html_tail;

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if (strncmp(path, "favicon.ico", 11) == 0) {
		struct stat st;
		if (stat("favicon.ico", &st) == 0)
			send_file(cfd, "./", "favicon.ico", false);
		else {
			std::string reply = http_200_header + "\r\n";

			if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
				log(LL_DEBUG, "short write on response header");
		}
	}
	else if (!allow_admin) {
		goto do404;
	}
	else if (strcmp(path, "pause") == 0 || strcmp(path, "unpause") == 0) {
		std::string reply = http_200_header + "???";

		std::string module;

		if (auto mod_it = pars.find("module"); mod_it != pars.end())
			module = mod_it -> second;

		cfg -> lock.lock();
		if (pause(inst, module, strcmp(path, "pause") == 0))
			reply = http_200_header + page_header + myformat(action_succeeded.c_str(), "pause/unpause") + html_tail;
		else
			reply = http_200_header + page_header + myformat(action_failed.c_str(), "pause/unpause") + html_tail;
		cfg -> lock.unlock();

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if (strcmp(path, "start") == 0 || strcmp(path, "stop") == 0) {
		std::string reply = http_200_header + "???";

		std::string module;

		if (auto mod_it = pars.find("module"); mod_it != pars.end())
			module = mod_it -> second;

		cfg -> lock.lock();
		interface *i = find_interface_by_id(inst, module);
		//printf("inst %s: %p\n", module.c_str(), i);

		if (!i)
			reply = "???";
		else if (start_stop(i, strcmp(path, "start") == 0))
			reply = http_200_header + page_header + myformat(action_succeeded.c_str(), "start/stop") + html_tail;
		else
			reply = http_200_header + page_header + myformat(action_failed.c_str(), "start/stop") + html_tail;
		cfg -> lock.unlock();

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if (strcmp(path, "restart") == 0) {
		std::string reply = http_200_header + "???";

		std::string module;

		if (auto mod_it = pars.find("module"); mod_it != pars.end())
			module = mod_it -> second;

		cfg -> lock.lock();
		interface *i = find_interface_by_id(inst, module);

		if (!i)
			reply = "???";
		else if (start_stop(i, false)) {
			if (start_stop(i, true))
				reply = http_200_header + page_header + myformat(action_succeeded.c_str(), "restart") + html_tail;
			else
				reply = http_200_header + page_header + myformat(action_failed.c_str(), "restart") + html_tail;
		}
		cfg -> lock.unlock();

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if (strcmp(path, "snapshot-img/") == 0) {
		std::string reply = http_200_header + "???";

		if (take_a_picture(s, snapshot_dir, quality, use_previous_when_fail))
			reply = http_200_header + page_header + myformat(action_succeeded.c_str(), "snapshot image") + html_tail;
		else
			reply = http_200_header + page_header + myformat(action_failed.c_str(), "snapshot image") + html_tail;

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else if (strcmp(path, "snapshot-video/") == 0) {
		interface *i = start_a_video(s, snapshot_dir, quality, cfg, is_view_proxy);
		//printf("interface %p instance %p\n", i, inst);

		std::string reply = http_200_header + "???";
		if (i) {
			reply = http_200_header + page_header + myformat(action_succeeded.c_str(), "snapshot video") + html_tail;

			cfg -> lock.lock();
			inst -> interfaces.push_back(i);
			cfg -> lock.unlock();
		}
		else {
			reply = http_200_header + page_header + myformat(action_failed.c_str(), "snapshot video") + html_tail;
		}

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response header");
	}
	else {
do404:
		log(LL_INFO, "Path %s not found", path);

		if (WRITE(cfd, "HTTP/1.0 404\r\n\r\nwhat?", 4) <= 0)
			log(LL_DEBUG, "short write on response header");
	}

	free(unescaped_path);
	free(Path);

	close(cfd);

	if (s)
		s -> unregister_user();
}

void * handle_http_client_thread(void *ct_in)
{
	http_thread_t *ct = (http_thread_t *)ct_in;

	set_thread_name("http_client");

	if (ct -> is_rest) {
		handle_rest(ct -> fd, ct -> cfg, ct -> global_stopflag, ct -> snapshot_dir, ct -> quality); // FIXME auth
	}
	else {
		handle_http_client(ct -> fd, ct -> fps, ct -> quality, ct -> time_limit, ct -> filters, ct -> global_stopflag, ct -> r, ct -> resize_w, ct -> resize_h, ct -> motion_compatible, ct -> snapshot_dir, ct -> allow_admin, ct -> archive_acces, ct -> cfg, ct -> limit_to, ct -> views, ct -> use_previous_when_fail, ct -> auth);
	}

	delete ct;

	return nullptr;
}

http_server::http_server(configuration_t *const cfg, http_auth *const auth, instance_t *const limit_to, const std::string & id, const std::string & descr, const std::string & http_adapter, const int http_port, const double fps, const int quality, const int time_limit, const std::vector<filter *> *const f, resize *const r, const int resize_w, const int resize_h, source *const motion_compatible, const bool allow_admin, const bool archive_acces, const std::string & snapshot_dir, const bool is_rest, instance_t  *const views, const bool use_previous_when_fail) : cfg(cfg), auth(auth), interface(id, descr), fps(fps), quality(quality), time_limit(time_limit), f(f), r(r), resize_w(resize_w), resize_h(resize_h), motion_compatible(motion_compatible), allow_admin(allow_admin), archive_acces(archive_acces), snapshot_dir(snapshot_dir), is_rest(is_rest), views(views), limit_to(limit_to), use_previous_when_fail(use_previous_when_fail)
{
	fd = start_listen(http_adapter.c_str(), http_port, 5);
	ct = CT_HTTPSERVER;
}

http_server::~http_server()
{
	close(fd);

	free_filters(f);

	delete f;
}

void http_server::operator()()
{
	set_thread_name("http_server");

	struct pollfd fds[] { { fd, POLLIN, 0 } };

	std::vector<pthread_t> handles;

	for(;!local_stop_flag;) {
		for(size_t i=0; i<handles.size();) {
			if (pthread_tryjoin_np(handles.at(i), nullptr) == 0)
				handles.erase(handles.begin() + i);
			else
				i++;
		}

		get_meta() -> set_int("$http-viewers$", std::pair<uint64_t, int>(0, handles.size()));

		pauseCheck();

		if (poll(fds, 1, 500) == 0)
			continue;

		int cfd = accept(fd, nullptr, 0);
		if (cfd == -1)
			continue;

		log(id, LL_INFO, "HTTP connected with: %s", get_endpoint_name(cfd).c_str());

		http_thread_t *ct = new http_thread_t;

		ct -> fd = cfd;
		ct -> fps = fps;
		ct -> quality = quality;
		ct -> time_limit = time_limit;
		ct -> filters = f;
		ct -> global_stopflag = &local_stop_flag;
		ct -> r = r;
		ct -> resize_w = resize_w;
		ct -> resize_h = resize_h;
		ct -> cfg = cfg;
		ct -> snapshot_dir = snapshot_dir;
		ct -> allow_admin = allow_admin;
		ct -> archive_acces = archive_acces;
		ct -> is_rest = is_rest;
		ct -> views = views;
		ct -> motion_compatible = motion_compatible;
		ct -> limit_to = limit_to;
		ct -> use_previous_when_fail = use_previous_when_fail;
		ct -> auth = auth;

		pthread_t th;
		int rc = -1;
		if ((rc = pthread_create(&th, nullptr, handle_http_client_thread, ct)) != 0)
		{
			errno = rc;
			error_exit(true, "pthread_create failed (http server)");
		}

		handles.push_back(th);
	}

	for(auto th : handles)
		pthread_join(th, nullptr);

	log(id, LL_INFO, "HTTP server thread terminating");
}
